@@ -118,8 +118,8 @@ def rotsph(lon,lat,clon,clat,anode=None,reverse=False,original=False):
     nlon = nlon % 360.0
 
     # Make negative points positive
-    bd, = np.where(nlon < 0.0)
-    if len(bd)>0:
+    bd = (nlon < 0.0)
+    if np.sum(bd)>0:
         nlon[bd] = nlon[bd]+360.0
 
     return nlon, nlat
@@ -269,6 +269,10 @@ def doPolygonsOverlap(xPolygon1, yPolygon1, xPolygon2, yPolygon2):
     n2 = len(xPolygon2)
     isin = False
 
+    # If ranges don't overlap, then polygons don't overlap
+    if rangeoverlap(xPolygon1,xPolygon2)==False or rangeoverlap(yPolygon1,yPolygon2)==False:
+        return False
+    
     # Loop through all vertices of second polygon
     for i in range(n2):
         # perform iterative boolean OR
@@ -279,6 +283,22 @@ def doPolygonsOverlap(xPolygon1, yPolygon1, xPolygon2, yPolygon2):
     for i in range(n1):
         isin = isin or isPointInPolygon(xPolygon2, yPolygon2, xPolygon1[i], yPolygon1[i])
 
+    # Two polygons can overlap even if there are no vertices inside each other.
+    # Need to check if the line segments overlap
+    if isin==False:
+        intersect = False
+        # Add first vertex to the end
+        xp1 = np.append( xPolygon1, xPolygon1[0] )
+        yp1 = np.append( yPolygon1, yPolygon1[0] )
+        xp2 = np.append( xPolygon2, xPolygon2[0] )
+        yp2 = np.append( yPolygon2, yPolygon2[0] )
+        for i in range(4):
+            for j in range(4):
+                intersect = intersect or doLineSegmentsIntersect(xp1[i:i+2],yp1[i:i+2],xp2[j:j+2],yp2[j:j+2])
+                if intersect==True:
+                    return True
+        isin = isin or intersect
+        
     return isin
 
 def isPointInPolygon(xPolygon, yPolygon, xPt, yPt):
@@ -333,7 +353,124 @@ def isLeft(x1, y1, x2, y2, x3, y3):
     # <0 for P3 right of the line
     return ( (x2 - x1) * (y3 - y1) - (x3 - x1) * (y2 - y1) )
 
+def rangeoverlap(a,b):
+    """does the range (start1, end1) overlap with (start2, end2)"""
+    return max(a) >= min(b) and min(a) <= max(b)
+    
+def doLineSegmentsIntersect(x1, y1, x2, y2):
+    """ Do two line segments intersect."""
 
+    # Check vertical lines
+
+    # Vertical lines, but NOT same X-values
+    if x1[0]==x1[1] and x2[0]==x2[1] and x1[0]!=x2[0]:
+        return False  # No overlap
+    
+    # Vertical lines with same X values
+    if x1[0]==x1[1] and x2[0]==x2[1] and x1[0]==x2[0]:
+        # Check intersection of Y ranges
+        I1 = [np.min(y1), np.max(y1)]
+        I2 = [np.min(y2), np.max(y2)]
+    
+        # And we could say that Xa is included into :
+        Ia = [max( np.min(y1), np.min(y2) ),
+              min( np.max(y1), np.max(y2) )]
+    
+        # Now, we need to check that this interval Ia exists :
+        if rangeoverlap(y1,y2)==False:
+            return False  # There is no mutual abcisses        
+        else:
+            return True   # There is overlap
+
+    # The equation of a line is:
+    #
+    # f(x) = A*x + b = y
+    # For a segment, it is exactly the same, except that x is included on an interval I.
+    # 
+    # If you have two segments, defined as follow:
+    #
+    # Segment1 = {(X1, Y1), (X2, Y2)}
+    # Segment2 = {(X3, Y3), (X4, Y4)}
+    # The abcisse Xa of the potential point of intersection (Xa,Ya) must be contained in both interval I1 and I2, defined as follow:
+    I1 = [np.min(x1), np.max(x1)]
+    I2 = [np.min(x2), np.max(x2)]
+    
+    # And we could say that Xa is included into :
+    Ia = [max( np.min(x1), np.min(x2) ),
+          min( np.max(x1), np.max(x2) )]
+    
+    # Now, we need to check that this interval Ia exists :
+    if rangeoverlap(x1,x2)==False:        
+        return False  # There is no mutual abcisses
+
+    # Check that the Y-ranges overlap as well
+    if rangeoverlap(y1,y2)==False:        
+        return False  # There is no mutual y-value overlap
+    
+    # So, we have two line formula, and a mutual interval. Your line formulas are:
+    # f1(x) = m1*x + b1 = y
+    # f2(x) = m2*x + b2 = y
+    
+    # As we got two points by segment, we are able to determine A1, A2, b1 and b2:
+    dx1 = x1[1]-x1[0]
+    if dx1==0:
+        m1 = np.inf
+        b1 = 0
+    else:
+        m1 = (y1[1]-y1[0])/(x1[1]-x1[0])  # Pay attention to not dividing by zero
+        b1 = y1[0]-m1*x1[0]        
+    dx2 = x2[1]-x2[0]
+    if dx2==0:
+        m2 = np.inf
+        b2 = 0
+    else:
+        m2 = (y2[1]-y2[0])/(x2[1]-x2[0])  # Pay attention to not dividing by zero
+        b2 = y2[0]-m2*x2[0]
+    
+    # If the segments are parallel, then m1 == m2:
+    if (m1 == m2) and (b1 != b2):
+        return False  # Parallel segments
+    
+    # If the segments are parallel and on top of each other, the m1==m2 and b1==b2
+    # we've already required that the x-ranges (abcissas) overlap
+    if (m1 == m2) and (b1 == b2):
+        return True   # parallel segments on top of each other
+    
+    # A point (Xa,Ya) standing on both lines must satisfy both formulas f1 and f2:
+    # Ya = m1 * Xa + b1
+    # Ya = m2 * Xa + b2
+    # A1 * Xa + b1 = m2 * Xa + b2
+
+    # Line segment 1 is vertical line 
+    if x1[0]==x1[1]:
+        Xa = x1[0]
+        Ya = m2*Xa+b2
+        if rangeoverlap(x1,[Xa]) and rangeoverlap(x2,[Xa]) and \
+           rangeoverlap(y1,[Ya]) and rangeoverlap(y2,[Ya]):
+            return True
+        else:
+            return False
+    # Line semgent 2 is vertical line
+    elif x2[0]==x2[1]:
+        Xa = x2[0]
+        Ya = m1*Xa+b1
+        if rangeoverlap(x1,[Xa]) and rangeoverlap(x2,[Xa]) and \
+           rangeoverlap(y1,[Ya]) and rangeoverlap(y2,[Ya]):        
+            return True
+        else:
+            return False
+    # Neither are vertical lines
+    else:
+        Xa = (b2 - b1) / (m1 - m2)
+        
+    # The last thing to do is check that Xa is included into Ia:
+    if ( (Xa < max( np.min(x1), np.min(x2) )) or
+         (Xa > min( np.max(x1), np.max(x2) )) ):
+        return False  # intersection is out of bound
+    else:
+        return True
+
+    
 # from astroML
 def crossmatch(X1, X2, max_distance=np.inf,k=1):
     """Cross-match the values between X1 and X2
@@ -655,7 +792,7 @@ class MagellanicStream(coord.BaseCoordinateFrame):
     radial_velocity : :class:`~astropy.units.Quantity`, optional, must be keyword
         The radial velocity of this object.
 
-    Developed by J. Povick
+    Developed by Y. Choi
 
     """
     default_representation = coord.SphericalRepresentation
