@@ -12,7 +12,9 @@ import re
 import logging
 import os
 import sys
+import time
 import numpy as np
+import shutil
 import warnings
 from astropy.io import fits
 from astropy.table import Table, Column
@@ -24,15 +26,20 @@ from scipy.stats import mstats
 from scipy.ndimage.filters import median_filter,gaussian_filter1d
 from scipy.optimize import curve_fit, least_squares
 from scipy.special import erf
-from scipy.interpolate import interp1d
+from scipy.interpolate import interp1d,splrep,BSpline
 from scipy.linalg import svd
 from scipy.signal import savgol_filter
 import astropy.stats
+from matplotlib.backend_bases import MouseButton
+import pandas as pd
+import matplotlib.pyplot as plt
+import traceback
 
 # Ignore these warnings, it's a bug
 warnings.filterwarnings("ignore", message="numpy.dtype size changed")
 warnings.filterwarnings("ignore", message="numpy.ufunc size changed")
 
+
 # NUMPY_LT_1_18
 def _expand_dims(data, axis):
     """
@@ -83,7 +90,10 @@ def _expand_dims(data, axis):
 def size(a=None):
     """Returns the number of elements"""
     if a is None: return 0
-    return np.array(a,ndmin=1).size
+    try:
+        return len(a)
+    except:
+        return np.array(a,ndmin=1).size
 
 
 # Median Absolute Deviation
@@ -140,8 +150,10 @@ def mad(data, axis=None, func=None, ignore_nan=True, zero=False):
 
     """
 
-    if type(data) is not np.ndarray: raise ValueError("data must be a numpy array")  
-
+    #if type(data) is not np.ndarray: raise ValueError("data must be a numpy array")  
+    if type(data) is not np.ndarray:
+        data = np.array(data)
+    
     if func is None:
         # Check if the array has a mask and if so use np.ma.median
         # See https://github.com/numpy/numpy/issues/7330 why using np.ma.median
@@ -192,13 +204,15 @@ def mad(data, axis=None, func=None, ignore_nan=True, zero=False):
 
 def minmax(a):
     """ Return a 2-element array of minimum and maximum."""
-    if type(a) is not np.ndarray: raise ValueError("a must be a numpy array")
+    if type(a) is not np.ndarray:
+        a = np.array(a)
     return np.array([np.min(a),np.max(a)])
 
 def stat(a=None,silent=False):
     """ Returns basic statistics on an array."""
     if a is None: raise ValueError("a must be input")
-    if type(a) is not np.ndarray: raise ValueError("a must be a numpy array")
+    if type(a) is not np.ndarray:
+        a = np.array(a)
     #  This is what stat returns:
     #  info[0]: Number of Elements
     #  info[1]: Minimum
@@ -260,6 +274,67 @@ def glob(inp):
         return out
     else:
         raise ValueError(inp,' not understood')
+
+def file_copy(src,dest,overwrite=False):
+    """ Copy files."""
+    nsrc = size(src)
+    if nsrc==1 and type(src) is str:
+        src = [src]
+    ndest = size(dest)
+    if ndest==1 and type(dest) is str:
+        dest = [dest]
+    # Loop over files
+    for i in range(nsrc):
+        src1 = src[i]
+        dest1 = dest[i]
+        if os.path.isdir(src1):
+            raise ValueError(src1+' is a directory')
+        # Destination is directory, add base name
+        if os.path.isdir(dest1):
+            dest1 += os.path.basename(src1)
+        # Check if the destination file exists already
+        exists1 = os.path.exists(dest1)
+        if os.path.abspath(src1)==os.path.abspath(dest1):
+            print(src1+' and '+dest1+' are the same file')
+            continue
+        if exists1 and overwrite==False:
+            print(dest1+' exists and overwrite=False')
+            continue
+        if exists1 and overwrite:
+            os.remove(exists1)
+        # Copy the file
+        shutil.copyfile(src1,dest1)
+
+def file_move(src,dest,overwrite=False):
+    """ Copy files.  dest can be a filename or directory."""
+    nsrc = size(src)
+    if nsrc==1 and type(src) is str:
+        src = [src]
+    ndest = size(dest)
+    if ndest==1 and type(dest) is str:
+        dest = [dest]
+    # Loop over files
+    for i in range(nsrc):
+        src1 = src[i]
+        dest1 = dest[i]
+        if os.path.isdir(src1):
+            raise ValueError(src1+' is a directory')
+        # Destination is directory, add base name
+        if os.path.isdir(dest1):
+            dest1 += os.path.basename(src1)
+        # Check if the destination file exists already
+        exists1 = os.path.exists(dest1)
+        if os.path.abspath(src1)==os.path.abspath(dest1):
+            print(src1+' and '+dest1+' are the same file')
+            continue
+        if exists1 and overwrite==False:
+            print(dest1+' exists and overwrite=False')
+            continue
+        if exists1 and overwrite:
+            os.remove(exists1)
+        # Move the file
+        shutil.move(src1,dest1)
+        
     
 def strlen(lst=None):
     """ Calculate the string lengths of a string array."""
@@ -381,7 +456,7 @@ def first_el(lst):
 # Standard grep function that works on string list
 def grep(lines=None,expr=None,index=False):
     """
-    Similar to the standard unit "grep" but run on a list of strings.
+    Similar to the standard unix "grep" but run on a list of strings.
     Returns a list of the matching lines unless index=True is set,
     then it returns the indices.
 
@@ -430,6 +505,43 @@ def grep(lines=None,expr=None,index=False):
         cnt = cnt+1
     return out
 
+# find() function for arrays
+def find(arr,expr):
+    """
+    The standard python string find() but for lists or np.char.array arrays.
+    Note that this uses re so it understands regular expressions.
+
+    Parameters
+    ----------
+    arr : list or np.char.array
+        The list of string lines to check.
+    expr  : str
+        Scalar string expression to search for.  Can be a regular expression.
+
+    Returns
+    -------
+    out : numpy array
+        The array of indices.
+
+    Example
+    -------
+
+    Search for a string and return the matching lines:
+
+    .. code-block:: python
+
+        mlines = find(arr,"hello")
+
+    """
+    out = []
+    for l in np.array(arr,ndmin=1):
+        m = re.search(expr,l)
+        if m != None:
+            out.append(m.start())
+        else:
+            out.append(-1)
+    return np.array(out).astype(int)
+
 # Create an empty file
 def touch(fname):
     open(fname, 'a').close()
@@ -913,6 +1025,21 @@ def quadratic_coefficients(x,y,axis=0):
         
     return coef
 
+def linear_coefficients(x,y,silent=True):
+    """ Calculate the lienar coefficients from two points."""
+
+    # y = mx + b
+
+    m = (y[1]-y[0])/(x[1]-x[0])
+    b = y[0] - m*x[0]
+
+    if silent==False:
+        if b<0:
+            print('y = '+str(m)+'*x - '+str(np.abs(b)))
+        else:
+            print('y = '+str(m)+'*x + '+str(b))            
+    return m,b
+    
 def wtmean(x,sigma,error=False,reweight=False,magnitude=False):
     """ Calculate weighted mean and error"""
     # np.average() can take weights as well.
@@ -1386,7 +1513,7 @@ def savgol(y,nbin,order=3):
     return yhat
         
 # Rebin data
-def rebin(arr, new_shape=None, binsize=None, tot=False,med=False):
+def rebin(arr, new_shape=None, binsize=None, tot=False,med=False,maximum=False):
     """
     Rebin data in 1D or 2D
 
@@ -1401,7 +1528,9 @@ def rebin(arr, new_shape=None, binsize=None, tot=False,med=False):
     tot : boolean, optional
        Return the sum instead of the mean (the default).
     med : boolean, optional
-       Return the median instead of the mean (the default).
+       Return the median instead of the mean.
+    maximum : boolean, optional
+       Return the maximum instead of the mean.
 
     Returns
     -------
@@ -1435,6 +1564,9 @@ def rebin(arr, new_shape=None, binsize=None, tot=False,med=False):
         # Sum
         elif tot:
             out = arr[slc].reshape(shape).sum(-1).sum(1)
+        # Maximum
+        elif maximum:
+            out = arr[slc].reshape(shape).max(-1).max(1)
         # Mean
         else:
             out = arr[slc].reshape(shape).mean(-1).mean(1)
@@ -1453,6 +1585,9 @@ def rebin(arr, new_shape=None, binsize=None, tot=False,med=False):
         # Sum
         elif tot:
             out = arr[slc].reshape(shape).sum(-1)
+        # Maximum
+        elif maximum:
+            out = arr[slc].reshape(shape).max(-1)        
         # Mean
         else:
             out = arr[slc].reshape(shape).mean(-1)
@@ -1851,18 +1986,6 @@ def onclick(event):
 
     return
 
-def clicker():
-    # im=fits.getdata('F1-00507803_23.fits.fz') 
-    # fig = plt.figure()
-    # plt.imshow(im)
-    # clicker()
-    # then the coordinates will be the global "coords" list
-    print('Click outside the plot to end')
-    global cid
-    cid = fig.canvas.mpl_connect('button_press_event', onclick)
-    coords = []
-
-
 def add_elements(cat,num=10000):
     """ Add more elements to a catalog"""
     ncat = len(cat)
@@ -2157,3 +2280,266 @@ def isnumber(s):
         return True
     except ValueError:
         return False
+
+def bootstrap(data,statistic,niter=100,indexargs=False,args=None,kwargs=None):
+    """
+    Perform bootstrap statistics.
+    
+    Parameters
+    ----------
+    data : list or numpy array
+       The data on which to perform the statistic.
+    statistic : str or callable
+       Name of statistical function ('sum','mean','median', etc.) or callable
+         function to compute statistic.
+    niter : int, optional
+       Number of bootstrap iterationas.  Default is 100.
+    indexargs : boolean, optional
+       Apply the random ordering to the args/kwargs that are arrays and have
+          the same size as data.  Default is False.
+    args : tuple, optional
+       Extra positional arguments to pass to `statistic`.
+    kwargs : dictionary, optional
+       Extra keyword arguments to pass to `statistic`.
+
+    Returns
+    -------
+    sigma : numpy array
+      The 1-sigma bootstrap values for each output that
+        statistic returns.
+
+    Example
+    -------
+
+    sigma = bootstrap(xdata,ydata,statistic)
+
+    """
+
+    # String name
+    if type(statistic) is str:
+        if statistic=='sum' or statistic=='np.sum' or statistic=='numpy.sum':
+            statistic = np.sum
+        if statistic=='nansum' or statistic=='np.nansum' or statistic=='numpy.nansum':
+            statistic = np.nansum
+        elif statistic=='max' or statistic=='np.max' or statistic=='numpy.max':
+            statistic = np.max
+        elif statistic=='nanmax' or statistic=='np.nanmax' or statistic=='numpy.nanmax':
+            statistic = np.nanmax
+        elif statistic=='min' or statistic=='np.min' or statistic=='numpy.min':
+            statistic = np.min
+        elif statistic=='nanmin' or statistic=='np.nanmin' or statistic=='numpy.nanmin':
+            statistic = np.nanmin                        
+        elif statistic=='mean' or statistic=='np.mean' or statistic=='numpy.mean':
+            statistic = np.mean
+        elif statistic=='nanmean' or statistic=='np.nanmean' or statistic=='numpy.nanmean':
+            statistic = np.nanmean            
+        elif statistic=='median' or statistic=='np.median' or statistic=='numpy.median':
+            statistic = np.median
+        elif statistic=='nanmedian' or statistic=='np.nanmedian' or statistic=='numpy.nanmedian':
+            statistic = np.nanmedian            
+        elif statistic=='std' or statistic=='np.std' or statistic=='numpy.std':
+            statistic = np.std
+        elif statistic=='nanstd' or statistic=='np.nanstd' or statistic=='numpy.nanstd':
+            statistic = np.nanstd            
+        elif statistic=='mad':
+            statistic = np.mad
+        else:
+            raise ValueError('Statistic '+statistic+' not supported')
+    ndata = np.array(data).size
+    # Get number of outputs    
+    vals = (data,)
+    if args is not None: vals = (vals,args)
+    if kwargs is not None: vals = (vals,kwargs)
+    out0 = statistic(*vals)
+    nout = np.array(out0).size
+
+    # Initialize seed using data so reproducible
+    if type(data[0]) is not int:
+        seed = int(np.abs(data[0])*1e5)
+    else:
+        seed = data[0]
+    rng = np.random.default_rng(seed)
+    # Bootstrap loop
+    ind = np.arange(ndata).astype(int)
+    btout = np.zeros((niter,nout),float)
+    for i in range(niter):
+        # Randomly pick elements with replacement
+        rndind = rng.choice(ind,ndata,replace=True)
+        data1 = data[rndind]   # apply to data array
+        vals = (data1,)
+        # Apply the random ordering to the input arguments if
+        #  they are arrays or lists
+        if indexargs:
+            # Apply random ordering to args
+            if args is not None:
+                if type(args) is not list and type(args) is not tuple:
+                    args = (args,)
+                nargs = len(args)
+                for j in range(nargs):
+                    # Array and same size as data, apply the random ordering
+                    if np.array(args[j]).size > 1 and np.array(args[j]).size==ndata:
+                        vals = (vals,args[j][rndind])
+                    else:
+                        vals = (vals,args[j])
+            # Apply random ordering to kwargs
+            if kwargs is not None:
+                nkwargs = len(kwargs)
+                kwargs1 = kwargs.copy()  # start new dictionary
+                for k in kwargs1.keys():
+                    # Array and same size as data, apply the random ordering                    
+                    if np.array(kwargs[k]).size > 1 and np.array(kwargs[k]).size==ndata:
+                        kwargs1[k] = kwargs[k][rndind]
+                vals = (vals,kwargs1)
+        # Do not apply random ordering to args/kwargs
+        else:
+            if args is not None: vals = (vals,args)
+            if kwargs is not None: vals = (vals,kwargs)
+        out1 = statistic(*vals)
+        btout[i,:] = out1
+    # Calculate robust standard deviation for each coefficient
+    sigma = np.zeros(nout,float)
+    for i in range(nout):
+        med = np.median(btout[:,i])
+        sigma[i] = mad(btout[:,i]-med,zero=True)
+    if nout==1:
+        sigma = sigma[0]
+        
+    return sigma
+
+def bspline(x,y,w=None,nquantiles=10,nord=3,knots=None,extrapolate=False):
+    """
+    Fit a B-Spline to data.  This is a thin wrapper for scipy.interpolate.splrep
+    and scipy.interpolate.BSpline.
+
+    Parameters
+    ----------
+    x : numpy array
+       Array of x-values to fit.
+    y : numpy array
+       Array of y-values to fit.
+    w : numpy array, optional
+       Array of weights for x/y.
+    nquantiles : int, optional
+       Number of quantiles to use to create the knots. Default is 10.
+    nord : int, optional
+       The degree of the spline fit.  Default is 3.
+    knots : numpy array, optional
+       The knot points for the BSpline.
+    extrapolate : bool, optional
+       Should the B-Spline be allowed to extrapolate.  Default is False.
+
+    Returns
+    -------
+
+    spline : BSpline object
+       BSpline object that can be used to interpolate
+
+
+    Example
+    -------
+
+    spline = bspline(x,y,knots=knots,nord=3)
+
+    """
+
+    # X must be unique and sorted
+    u,ui = np.unique(x,return_index=True)
+    si = ui[np.argsort(x[ui])]
+    # Knots
+    if knots is None and nquantiles is None:
+        raise ValueError('nquantiles or knots must be input')
+    if knots is None:
+        qs = np.linspace(0, 1, nquantiles+2)[1:-1]
+        knots = np.quantile(x[si], qs)
+    # Perform the fitting
+    if w is not None:
+        tck = splrep(x[si],y[si],w[si], t=knots, k=nord)
+    else:
+        tck = splrep(x[si],y[si], t=knots, k=nord)
+    # Generate the BSpline object
+    spline = BSpline(*tck, extrapolate=extrapolate)
+
+    return spline
+
+
+def help(*args,verbose=False):
+    """ Like IDL help"""
+
+    if len(args)==0:
+        locals()
+        
+    # Get variables names input
+    stack = traceback.extract_stack()
+    filename, lineno, function_name, code = stack[-2]
+    varnames = code[code.find('(')+1:-1].split(',')
+    varnames = [v.strip() for v in varnames]
+    
+    # Loop over arguments
+    for i in range(len(args)):
+        a = args[i]
+        atype = type(a)
+        try:
+            ndim = a.ndim
+        except:
+            ndim = 0
+        try:
+            shape = a.shape
+        except:
+            shape = None
+        try:
+            na = len(a)
+        except:
+            na = 0
+            
+        # Is this a table?
+        tab = False
+        if atype == Table:
+            colnames = a.colnames
+            dtype = a.dtype
+            tab = True
+        elif atype == np.ndarray and a.dtype.names is not None:
+            colnames = a.columns
+            dtype = a.dtype
+            tab = True
+        elif atype == pd.DataFrame:
+            colnames = a.columns
+            dtype = a.dtype
+            tab = True
+            
+        # Short output
+        if verbose==False or tab==False:
+            fmt = '%-16s %-22s %-10s'
+            if atype == np.ndarray and a.dtype.names is None:
+                atype = '<np.ndarray '+str(a.dtype)+'>'
+            if ndim==0:
+                data = (varnames[i],atype,'scalar')
+            elif shape is not None:
+                if len(shape)==1:
+                    sshape = '['+str(shape[0])+']'
+                else:
+                    sshape = '['+','.join(np.char.array(shape).astype(str))+']'
+                data = (varnames[i],atype,sshape)
+            else:
+                data = (varnames[i],atype,'['+str(na)+']')
+            if len(varnames[i])>16:
+                print(varnames[i])
+                print(fmt % ('',data[1],data[2]))
+            else:
+                print(fmt % data)
+
+        # Verbose output of table
+        else:
+            print(str(type(a))+', '+str(len(colnames))+' columns, ['+str(na)+']')
+            fmt = '%-20s %-10s %-20s'                    
+            for j in range(len(colnames)):
+                col = a[colnames[j]][0]
+                if size(col)>1:
+                    data = (colnames[j],dtype[j].base,'Array['+str(size(col))+']')           
+                else:
+                    data = (colnames[j],dtype[j].base,str(col))
+                if len(colnames[j])>16:
+                    print(colnames[j])
+                    print(fmt % ('',data[1],data[2]))
+                else:
+                    print(fmt % data)
+
